import logging
from collections.abc import Sequence

import loguru

from ._icecream import init_icecream
from ._init_loguru import init_loguru


def init_logging(
    level: int | str = logging.NOTSET,
    *,
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
) -> None:
    init_loguru(level=level, handlers=handlers, levels=levels)
    init_icecream()



import builtins
import functools
from collections.abc import Callable, Mapping

import loguru
from loguru import _filters, logger

from liblaf import grapes

from .typed import Filter

dispatcher = grapes.ConditionalDispatcher()


# ref: <https://github.com/Delgan/loguru/blob/master/loguru/_logger.py#L259>


@dispatcher.register(lambda filter_: filter_ is None)
def _(filter_: None) -> None:  # noqa: ARG001
    return None


@dispatcher.register(lambda filter_: filter_ == "")
def _(filter_: str) -> "loguru.FilterFunction":  # noqa: ARG001
    return _filters.filter_none


@dispatcher.register(lambda filter_: isinstance(filter_, str))
def _(filter_: str) -> "loguru.FilterFunction":
    parent: str = filter_ + "."
    length: int = len(parent)
    return functools.partial(_filters.filter_by_name, parent=parent, length=length)


@dispatcher.register(lambda filter_: isinstance(filter_, Mapping))
def _(filter_: Mapping[str, int | str]) -> "loguru.FilterFunction":
    level_per_module: dict[str, int] = {}
    for module, level_ in filter_.items():
        if module is not None and not isinstance(module, str):
            msg: str = (
                "The filter dict contains an invalid module, "
                f"it should be a string (or None), not: '{type(module).__name__}'"
            )
            raise TypeError(msg)
        if level_ is False:
            levelno_ = False
        elif level_ is True:
            levelno_ = 0
        elif isinstance(level_, str):
            try:
                levelno_: int = logger.level(level_).no
            except ValueError:
                msg = (
                    f"The filter dict contains a module '{module}' associated to a level name "
                    f"which does not exist: '{level_}'"
                )
                raise ValueError(msg) from None
        elif isinstance(level_, int):
            levelno_ = level_
        else:
            msg = (
                f"The filter dict contains a module '{module}' associated to an invalid level, "
                f"it should be an integer, a string or a boolean, not: '{type(level_).__name__}'"
            )
            raise TypeError(msg)
        if levelno_ < 0:
            msg = (
                f"The filter dict contains a module '{module}' associated to an invalid level, "
                f"it should be a positive integer, not: '{levelno_}'"
            )
            raise ValueError(msg)
        level_per_module[module] = levelno_
    return functools.partial(
        _filters.filter_by_level, level_per_module=level_per_module
    )


@dispatcher.register(callable)
def _(filter_: Callable) -> "loguru.FilterFunction":
    if filter_ == builtins.filter:
        msg = (
            "The built-in 'filter()' function cannot be used as a 'filter' parameter, "
            "this is most likely a mistake (please double-check the arguments passed "
            "to 'logger.add()')."
        )
        raise ValueError(msg)
    return filter_


@dispatcher.final(fallback=True)
def as_filter_func(filter_: Filter) -> "loguru.FilterFunction | None":
    msg: str = f"Invalid filter, it should be a function, a string or a dict, not: '{type(filter_).__name__}'"
    raise TypeError(msg)



"""This module provides functions for serialization and deserialization of various data formats, including JSON, TOML, YAML, and Pydantic models. It also includes registries for mapping file extensions to their respective serialization and deserialization functions."""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)



from collections.abc import Callable
from pathlib import Path

from rich.style import Style
from rich.text import Text


def func(obj: Callable) -> Text:
    text = Text()
    file: Path = Path(obj.__code__.co_filename)
    if file.exists():
        text.append(obj.__module__, style=Style(link=file.as_uri()))
        text.append(".")
        text.append(
            f"{obj.__qualname__}(...)",
            style=Style(link=f"{file.as_uri()}#{obj.__code__.co_firstlineno}"),
        )
    else:
        text.append(f"{obj.__module__}.{obj.__qualname__}(...)")
    return text



from collections.abc import Generator, Iterable, Sequence

from rich.progress import Progress

from liblaf import grapes

from ._progress import progress


def track[T](
    iterable: Iterable[T],
    *,
    description: str | bool | None = True,
    log_level_record: int | str | None = "DEBUG",
    log_level_summary: int | str | None = "INFO",
    timers: bool | Sequence[grapes.TimerName | str] = ["perf"],
    total: float | None = None,
) -> Generator[T]:
    if description is True:
        description = grapes.caller_location(2).plain
    description = description or ""
    prog: Progress = progress()
    if timers is True:
        timers = ["perf"]
    if total is None:
        total = try_len(iterable)
    if timers:
        iterable: grapes.TimedIterable[T] = grapes.timer(
            iterable,
            depth=2,
            label=description,
            log_level_record=log_level_record,
            log_level_summary=log_level_summary,
            timers=timers,
            total=int(total) if total is not None else None,
        )
        with prog:
            yield from prog.track(iterable, total=total, description=description)
    else:
        with prog:
            yield from prog.track(iterable, total=total, description=description)


def try_len(iterable: Iterable) -> int | None:
    try:
        return len(iterable)  # pyright: ignore[reportArgumentType]
    except TypeError:
        return None



import functools
from collections.abc import Callable

import attrs

from liblaf import grapes

from ._base import TimerRecords


# `slots=False` is required to make `functools.update_wrapper(...)` work
# ref: <https://www.attrs.org/en/stable/glossary.html#term-slotted-classes>
@attrs.define(slots=False)
class TimedFunction[**P, T](TimerRecords):
    _func: Callable[P, T] = attrs.field(alias="func", on_setattr=attrs.setters.frozen)

    def __attrs_post_init__(self) -> None:
        self.label = self.label or grapes.pretty.func(self._func).plain or "Function"
        functools.update_wrapper(self, self._func)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        self._start()
        result: T = self._func(*args, **kwargs)
        self._end()
        self.log_record(depth=2)
        return result



import os
from pathlib import Path

from rich.style import Style
from rich.text import Text


def location(
    name: str | None,
    function: str | None,
    line: int | None,
    file: str | os.PathLike[str] | None = None,
) -> Text:
    text = Text()
    file: Path | None = Path(file or "<unknown>")
    function = function or "<unknown>"
    line = line or 0
    if file.exists():
        text.append(str(name), style=Style(link=file.as_uri()))
        text.append(":")
        text.append(f"{function}:{line}", style=Style(link=f"{file.as_uri()}#{line}"))
    else:
        text.append(f"{name}:{function}:{line}")
    return text



import os
import time
from typing import Literal

import autoregistry

type TimerName = Literal[
    "children_system",
    "children_user",
    "elapsed",
    "monotonic",
    "perf",
    "process",
    "system",
    "thread",
    "time",
    "user",
]


REGISTRY = autoregistry.Registry()
REGISTRY["children_system"] = lambda: os.times().children_system
REGISTRY["children_user"] = lambda: os.times().children_user
REGISTRY["elapsed"] = lambda: os.times().elapsed
REGISTRY["monotonic"] = time.monotonic
REGISTRY["perf"] = time.perf_counter
REGISTRY["process"] = time.process_time
REGISTRY["system"] = lambda: os.times().system
REGISTRY["thread"] = time.thread_time
REGISTRY["time"] = time.time
REGISTRY["user"] = lambda: os.times().user


def get_time(name: TimerName | str = "perf") -> float:
    return REGISTRY[name]()



import os
from pathlib import Path
from typing import Any, override

import autoregistry
from typing_extensions import deprecated

from liblaf import grapes

from ._abc import AbstractSerializer
from ._json import json
from ._toml import toml
from ._yaml import yaml

SERIALIZERS = autoregistry.Registry()
SERIALIZERS["json"] = json
SERIALIZERS["toml"] = toml
SERIALIZERS["yaml"] = yaml
SERIALIZERS["yml"] = yaml


class AutoSerializer(AbstractSerializer):
    @override
    def load(
        self, fpath: str | os.PathLike[str], *, ext: str | None = None, **kwargs
    ) -> Any:
        serializer: AbstractSerializer = self.get_serializer(fpath, ext=ext)
        return serializer.load(fpath, **kwargs)

    @override
    def loads(self, data: str, *, ext: str | None = None, **kwargs) -> Any:
        serializer: AbstractSerializer = self.get_serializer(fpath="", ext=ext)
        return serializer.loads(data, **kwargs)

    @override
    def save(
        self,
        fpath: str | os.PathLike[str],
        data: Any,
        *,
        ext: str | None = None,
        **kwargs,
    ) -> None:
        serializer: AbstractSerializer = self.get_serializer(fpath, ext=ext)
        serializer.save(fpath, data, **kwargs)

    @override
    def saves(self, data: Any, *, ext: str | None = None, **kwargs) -> str:
        serializer: AbstractSerializer = self.get_serializer(fpath="", ext=ext)
        return serializer.saves(data, **kwargs)

    def get_serializer(
        self, fpath: str | os.PathLike[str], *, ext: str | None = None
    ) -> AbstractSerializer:
        if ext is None:
            fpath: Path = grapes.as_path(fpath)
            ext = fpath.suffix.lstrip(".")
        return SERIALIZERS[ext]  # pyright: ignore[reportReturnType]


auto = AutoSerializer()
load = auto.load
loads = auto.loads
save = auto.save
saves = auto.saves


@deprecated("Use `save()` instead of `serialize()`")
def serialize(
    fpath: str | os.PathLike[str], data: Any, *, ext: str | None = None
) -> None:
    return save(fpath, data, ext=ext)


@deprecated("Use `load()` instead of `deserialize()`")
def deserialize(fpath: str | os.PathLike[str], *, ext: str | None = None) -> Any:
    return load(fpath, ext=ext)



from . import typed
from ._as_filter_func import as_filter_func
from ._composite import filter_all, filter_any
from ._once import filter_once
from .typed import Filter

__all__ = [
    "Filter",
    "as_filter_func",
    "filter_all",
    "filter_any",
    "filter_once",
    "typed",
]



import contextlib
import types
from collections.abc import Callable, Iterable, Sequence
from typing import Self, overload

import attrs

from liblaf.grapes.timing._time import TimerName

from ._base import TimerRecords
from ._function import TimedFunction
from ._iterable import TimedIterable


@attrs.define
class Timer(
    contextlib.AbstractAsyncContextManager,
    contextlib.AbstractContextManager,
    TimerRecords,
):
    async def __aenter__(self) -> Self:
        return self.__enter__()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
        /,
    ) -> None:
        return self.__exit__(exc_type, exc_value, traceback)

    def __call__[**P, T](self, func: Callable[P, T]) -> TimedFunction[P, T]:
        return TimedFunction(
            func,
            label=self.label,
            timers=self.timers,
            log_level_record=self.log_level_record,
            log_level_summary=self.log_level_summary,
            log_summary_at_exit=self.log_summary_at_exit,
        )

    def __enter__(self) -> Self:
        self._start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
        /,
    ) -> None:
        self._end()
        self.log_record(depth=2)

    def start(self) -> None:
        self._start()

    def end(self, depth: int = 2) -> None:
        self._end()
        self.log_record(depth=depth)


@overload
def timer[T](
    iterable: Iterable[T],
    *,
    depth: int = 0,
    label: str | None = None,
    log_level_record: int | str | None = "DEBUG",
    log_level_summary: int | str | None = "INFO",
    log_summary_at_exit: bool = False,
    timers: Sequence[TimerName | str] = ["perf"],
    total: int | None = None,
) -> TimedIterable[T]: ...
@overload
def timer(
    iterable: None = None,
    *,
    depth: int = 0,
    label: str | None = None,
    log_level_record: int | str | None = "DEBUG",
    log_level_summary: int | str | None = "INFO",
    log_summary_at_exit: bool = False,
    timers: Sequence[TimerName | str] = ["perf"],
    total: None = None,
) -> Timer: ...
def timer[T](
    iterable: Iterable[T] | None = None,
    *,
    depth: int = 0,
    label: str | None = None,
    log_level_record: int | str | None = "DEBUG",
    log_level_summary: int | str | None = "INFO",
    log_summary_at_exit: bool = False,
    timers: Sequence[TimerName | str] = ["perf"],
    total: int | None = None,
) -> TimedIterable[T] | Timer:
    if iterable is not None:
        return TimedIterable(
            iterable=iterable,
            label=label,
            timers=timers,
            depth=depth,
            log_level_record=log_level_record,
            log_level_summary=log_level_summary,
            log_summary_at_exit=log_summary_at_exit,
            total=total,
        )
    return Timer(
        label=label,
        timers=timers,
        depth=depth,
        log_level_record=log_level_record,
        log_level_summary=log_level_summary,
        log_summary_at_exit=log_summary_at_exit,
    )



from collections.abc import Sequence

import loguru

from .filter_._composite import filter_all
from .filter_._once import filter_once
from .filter_.typed import Filter

DEFAULT_LEVEL: int | str = "DEBUG"


def default_filter() -> Filter:
    return filter_all(
        {
            "": "INFO",
            "__main__": "TRACE",
            "liblaf": "DEBUG",
        },
        filter_once(),
    )


DEFAULT_LEVELS: Sequence["loguru.LevelConfig"] = [
    {"name": "ICECREAM", "no": 15, "color": "<magenta><bold>", "icon": "🍦"}
]



import json as json_
import os
from pathlib import Path
from typing import Any, override

from liblaf import grapes

from ._abc import AbstractSerializer


class JSONSerializer(AbstractSerializer):
    @override
    def load(self, fpath: str | os.PathLike[str], **kwargs) -> Any:
        fpath: Path = grapes.as_path(fpath)
        with fpath.open() as fp:
            return json_.load(fp, **kwargs)

    @override
    def loads(self, data: str, **kwargs) -> Any:
        return json_.loads(data, **kwargs)

    @override
    def save(self, fpath: str | os.PathLike[str], data: Any, **kwargs) -> None:
        fpath: Path = grapes.as_path(fpath)
        with fpath.open("w") as fp:
            json_.dump(data, fp, **kwargs)

    @override
    def saves(self, data: Any, **kwargs) -> str:
        return json_.dumps(data, **kwargs)


json = JSONSerializer()
load_json = json.load
loads_json = json.loads
save_json = json.save
saves_json = json.saves



import types
from typing import Self, TypedDict

import attrs
import loguru
from rich.console import Console, RenderableType
from rich.highlighter import Highlighter, ReprHighlighter
from rich.table import Table
from rich.text import Text
from rich.traceback import Traceback

from liblaf import grapes


class TracebackArgs(TypedDict, total=False):
    show_locals: bool


@attrs.frozen(kw_only=True)
class RichLogRecordRenderer:
    console: Console = attrs.field(factory=grapes.logging_console)
    highlighter: Highlighter = attrs.field(factory=ReprHighlighter)
    markup: bool = attrs.field(default=True)
    traceback: TracebackArgs = attrs.field(
        factory=lambda: TracebackArgs(show_locals=True)
    )

    row: list[RenderableType] = attrs.field(init=False, factory=list)
    table: Table = attrs.field(
        init=False, factory=lambda: Table.grid(padding=(0, 1), expand=True)
    )

    def add_time(self, record: "loguru.Record") -> Self:
        self.table.add_column("Time", style="log.time")
        self.row.append(str(record["elapsed"]))
        return self

    def add_level(self, record: "loguru.Record") -> Self:
        self.table.add_column("Level", style="log.level", width=8)
        level: str = record["level"].name
        self.row.append(Text(level, style=f"logging.level.{level.lower()}"))
        return self

    def add_message(self, record: "loguru.Record") -> Self:
        self.table.add_column("Message", style="log.message", ratio=1)
        if (rich := record["extra"].get("rich")) is not None:
            self.row.append(rich)
            return self
        message: RenderableType = record["message"].rstrip()
        if record["extra"].get("markup", self.markup):
            message = Text.from_markup(message)
        if highlighter := record["extra"].get("highlighter", self.highlighter):
            message = highlighter(message)
        self.row.append(message)
        return self

    def add_path(self, record: "loguru.Record") -> Self:
        self.table.add_column("Path", style="log.path")
        path: Text = grapes.pretty.location(
            name=record["name"],
            function=record["function"],
            line=record["line"],
            file=record["file"].path,
        )
        self.row.append(path)
        return self

    def render_exception(self, record: "loguru.Record") -> RenderableType | None:
        exception: loguru.RecordException | None = record["exception"]
        if exception is None:
            return None
        exc_type: type[BaseException] | None
        exc_value: BaseException | None
        traceback: types.TracebackType | None
        exc_type, exc_value, traceback = exception
        if exc_type is None or exc_value is None:
            return None
        return Traceback.from_exception(
            exc_type=exc_type,
            exc_value=exc_value,
            traceback=traceback,
            width=self.console.width,
            code_width=self.console.width,
            **self.traceback,
        )

    def render(self) -> RenderableType:
        self.table.add_row(*self.row)
        return self.table


@attrs.define(kw_only=True)
class LoguruRichHandler:
    console: Console = attrs.field(factory=grapes.logging_console)
    highlighter: Highlighter = attrs.field(factory=ReprHighlighter)
    markup: bool = attrs.field(default=True)
    traceback: TracebackArgs = attrs.field(
        factory=lambda: TracebackArgs(show_locals=True)
    )

    def __call__(self, message: "loguru.Message") -> None:
        renderer = RichLogRecordRenderer(
            console=self.console,
            highlighter=self.highlighter,
            markup=self.markup,
            traceback=self.traceback,
        )
        renderer.add_time(message.record)
        renderer.add_level(message.record)
        renderer.add_message(message.record)
        renderer.add_path(message.record)
        self.console.print(renderer.render())
        if (exception := renderer.render_exception(message.record)) is not None:
            self.console.print(exception)



import sys
from pathlib import Path


def find_project_dir(start: Path | None = None, name: str = "src") -> Path:
    if start is None:
        start = Path(sys.argv[0])
    start = start.absolute()
    path: Path = Path(start)
    while path.name != name:
        if path.parent == path:
            return start.parent
        path = path.parent
    return path.parent



import os
from pathlib import Path
from typing import Unpack

import loguru
from environs import env
from rich.console import Console
from typing_extensions import deprecated

from liblaf import grapes

from ._default import default_filter
from .filter_ import Filter
from .handler import LoguruRichHandler, TracebackArgs


@deprecated("Use `rich_handler()` instead.")
def console_handler(
    console: Console | None = None,
    filter_: Filter | None = None,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> "loguru.HandlerConfig":
    if console is None:
        console = grapes.logging_console()
    if filter_ is None:
        filter_ = default_filter()

    def sink(message: "loguru.Message") -> None:
        console.print(message, end="", no_wrap=True, crop=False, overflow="ignore")

    return {
        "sink": sink,
        "format": "[green]{time:YYYY-MM-DD HH:mm:ss.SSS}[/green] | [logging.level.{level}]{level: <8}[/logging.level.{level}] | [cyan]{name}[/cyan]:[cyan]{function}[/cyan]:[cyan]{line}[/cyan] - {message}",
        "filter": filter_,
        **kwargs,
    }


def file_handler(
    fpath: str | os.PathLike[str] | None = None,
    filter_: Filter | None = None,
    **kwargs: Unpack["loguru.FileHandlerConfig"],
) -> "loguru.HandlerConfig":
    if fpath is None:
        fpath = env.path("LOGGING_FILE", default=Path("run.log"))
    if filter_ is None:
        filter_ = default_filter()
    if "format" not in kwargs:
        kwargs["format"] = (
            "<green>{elapsed}</green> | "
            "<level>{level: <8}</level> | "
            "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
            "<level>{message}</level>"
        )
    return {"sink": fpath, "filter": filter_, "mode": "w", **kwargs}


def jsonl_handler(
    fpath: str | os.PathLike[str] | None = None,
    filter_: Filter | None = None,
    **kwargs: Unpack["loguru.FileHandlerConfig"],
) -> "loguru.HandlerConfig":
    if fpath is None:
        fpath = env.path("LOGGING_JSONL", default=Path("run.log.jsonl"))
    if filter_ is None:
        filter_ = default_filter()
    return {"sink": fpath, "filter": filter_, "serialize": True, "mode": "w", **kwargs}


def rich_handler(
    *,
    console: Console | None = None,
    filter_: Filter | None = None,
    traceback: TracebackArgs | None = None,
    **kwargs: Unpack["loguru.BasicHandlerConfig"],
) -> "loguru.HandlerConfig":
    if console is None:
        console = grapes.logging_console()
    if filter_ is None:
        filter_ = default_filter()
    if traceback is None:
        traceback = TracebackArgs(show_locals=True)
    return {
        "sink": LoguruRichHandler(console=console, traceback=traceback),
        "format": "",
        "filter": filter_,
        **kwargs,
    }



from collections.abc import Callable, Hashable, Sequence

import glom
import loguru

DEFAULT_SPECS: Sequence[str] = (
    "file.path",
    "function",
    "level.no",
    "line",
    "message",
    "module",
    "name",
)


def default_transform(record: "loguru.Record") -> Hashable:
    return tuple(glom.glom(record, spec) for spec in DEFAULT_SPECS)


def filter_once(
    transform: "Callable[[loguru.Record], Hashable]" = default_transform,
) -> "loguru.FilterFunction":
    history: set[Hashable] = set()

    def filter_(record: "loguru.Record") -> bool:
        if not record["extra"].get("once", False):
            return True
        transformed: Hashable = transform(record)
        if transformed in history:
            return False
        history.add(transformed)
        return True

    return filter_



"""This module provides utility functions for handling optional imports and checking module availability."""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)



import contextlib

from loguru import logger


def add_level(
    name: str, no: int, color: str | None = None, icon: str | None = None
) -> None:
    """Add a new logging level to the logger.

    Args:
        name: The name of the new logging level.
        no: The numeric value of the new logging level.
        color: The color associated with the new logging level.
        icon: The icon associated with the new logging level.
    """
    with contextlib.suppress(ValueError):
        logger.level(name, no, color=color, icon=icon)



import logging
from collections.abc import Sequence

import loguru
from loguru import logger

from ._clear_handlers import clear_handlers
from ._default import DEFAULT_LEVELS
from ._handler import file_handler, jsonl_handler, rich_handler
from ._intercept import setup_loguru_logging_intercept
from ._level import add_level
from .filter_ import Filter


def init_loguru(
    level: int | str = logging.NOTSET,
    filter_: Filter | None = None,
    handlers: Sequence["loguru.HandlerConfig"] | None = None,
    levels: Sequence["loguru.LevelConfig"] | None = None,
) -> None:
    """Initialize the Loguru logger with specified configurations.

    Args:
        level: The logging level.
        filter_: A filter to apply to the logger.
        handlers: A sequence of handler configurations.
        levels: A sequence of level configurations.
    """
    if handlers is None:
        handlers: list[loguru.HandlerConfig] = [
            rich_handler(level=level, filter_=filter_)
        ]
        handlers.append(file_handler(level=level, filter_=filter_))
        handlers.append(jsonl_handler(level=level, filter_=filter_))
    logger.configure(handlers=handlers)
    for lvl in levels or DEFAULT_LEVELS:
        add_level(**lvl)
    setup_loguru_logging_intercept(level=level)
    clear_handlers()



from . import filter_, handler
from ._caller import caller_location
from ._clear_handlers import clear_handlers
from ._default import DEFAULT_LEVEL, DEFAULT_LEVELS, default_filter
from ._handler import console_handler, file_handler, jsonl_handler, rich_handler
from ._icecream import init_icecream
from ._init import init_logging
from ._init_loguru import init_loguru
from ._intercept import InterceptHandler, setup_loguru_logging_intercept
from ._level import add_level
from ._once import (
    critical_once,
    debug_once,
    error_once,
    exception_once,
    info_once,
    log_once,
    success_once,
    trace_once,
    warning_once,
)
from ._rich import init_rich, logging_console, logging_theme
from .filter_ import Filter, as_filter_func, filter_all, filter_any, filter_once
from .handler import LoguruRichHandler, TracebackArgs

__all__ = [
    "DEFAULT_LEVEL",
    "DEFAULT_LEVELS",
    "Filter",
    "InterceptHandler",
    "LoguruRichHandler",
    "TracebackArgs",
    "add_level",
    "as_filter_func",
    "caller_location",
    "clear_handlers",
    "console_handler",
    "critical_once",
    "debug_once",
    "default_filter",
    "error_once",
    "exception_once",
    "file_handler",
    "filter_",
    "filter_all",
    "filter_any",
    "filter_once",
    "handler",
    "info_once",
    "init_icecream",
    "init_logging",
    "init_loguru",
    "init_rich",
    "jsonl_handler",
    "log_once",
    "logging_console",
    "logging_theme",
    "rich_handler",
    "setup_loguru_logging_intercept",
    "success_once",
    "trace_once",
    "warning_once",
]



import os
import types
from typing import Literal

type ClassInfo = type | types.UnionType | tuple[ClassInfo, ...]
type LogLevel = (
    Literal["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"]  # noqa: PYI051
    | int
    | str
)
type StrPath = str | os.PathLike[str]



import bisect
import functools
import operator
from collections.abc import Callable
from typing import Any, Literal, NamedTuple, NoReturn

from liblaf.grapes.typed._decorator import Decorator


class Function(NamedTuple):
    condition: Callable[..., bool]
    function: Callable
    priority: int


class NotFoundLookupError(LookupError): ...


def _always_true(*args, **kwargs) -> Literal[True]:  # noqa: ARG001
    return True


def _dummy_final(*args, **kwargs) -> NoReturn:  # noqa: ARG001
    msg: str = "No condition matched."
    raise NotFoundLookupError(msg)


class ConditionalDispatcher:
    registry: list[Function]
    _final: Callable

    def __init__(self) -> None:
        self.registry = []
        self._final = _dummy_final

    def __call__(self, *args, **kwargs) -> Any:
        for func in self.registry:
            try:
                if func.condition(*args, **kwargs):
                    return func.function(*args, **kwargs)
            except TypeError:
                continue
        return self._final(*args, **kwargs)

    def final(self, /, *, fallback: bool = False) -> Decorator:
        def decorator[**P, T](func: Callable[P, T]) -> Callable[P, T]:
            if fallback:
                self._final = func
            else:
                self._final = _dummy_final
            functools.update_wrapper(self, func)
            return self

        return decorator

    def register(
        self, condition: Callable[..., bool] = _always_true, *, priority: int = 0
    ) -> Decorator:
        def decorator[**P, T](fn: Callable[P, T]) -> Callable[P, T]:
            bisect.insort(
                self.registry,
                Function(condition=condition, function=fn, priority=priority),
                key=operator.attrgetter("priority"),
            )
            functools.update_wrapper(self, fn)
            return self

        return decorator



from types import FrameType

from loguru._get_frame import get_frame
from rich.text import Text

from liblaf import grapes


def caller_location(depth: int = 1) -> Text:
    """Returns the file name and line number of the caller's location in the code.

    Args:
        depth: The stack depth to inspect.
        markup: If `True`, returns the file name and line number with markup for links.

    Returns:
        The file name and line number of the caller's location. If the frame cannot be retrieved, returns "<unknown>".
    """
    frame: FrameType | None
    try:
        frame = get_frame(depth)  # pyright: ignore[reportAssignmentType]
    except ValueError:
        frame = None
    file: str | None = None
    function: str | None = None
    line: int | None = None
    name: str | None = None
    if frame is not None:
        file = frame.f_code.co_filename
        function = frame.f_code.co_name
        line = frame.f_lineno
        name = frame.f_globals.get("__name__")
    text: Text = grapes.pretty.location(
        function=function, line=line, name=name, file=file
    )
    return text



from os import PathLike
from pathlib import Path


def as_path(path: str | PathLike[str], *, expend_user: bool = True) -> Path:
    path = Path(path)
    if expend_user:
        path = path.expanduser()
    return path



from . import conditional_dispatch
from ._decorator import decorator_with_optional_arguments
from .conditional_dispatch import ConditionalDispatcher

__all__ = [
    "ConditionalDispatcher",
    "conditional_dispatch",
    "decorator_with_optional_arguments",
]



from ._abc import AbstractSerializer
from ._json import (
    JSONSerializer,
    json,
    load_json,
    loads_json,
    save_json,
    saves_json,
)
from ._pydantic import load_pydantic, loads_pydantic, save_pydantic, saves_pydantic
from ._serde import (
    AutoSerializer,
    auto,
    deserialize,
    load,
    loads,
    save,
    saves,
    serialize,
)
from ._toml import (
    TOMLSerializer,
    load_toml,
    loads_toml,
    save_toml,
    saves_toml,
    toml,
)
from ._yaml import (
    YAMLSerializer,
    load_yaml,
    loads_yaml,
    save_yaml,
    saves_yaml,
    yaml,
)

__all__ = [
    "AbstractSerializer",
    "AutoSerializer",
    "JSONSerializer",
    "TOMLSerializer",
    "YAMLSerializer",
    "auto",
    "deserialize",
    "json",
    "load",
    "load_json",
    "load_pydantic",
    "load_toml",
    "load_yaml",
    "loads",
    "loads_json",
    "loads_pydantic",
    "loads_toml",
    "loads_yaml",
    "save",
    "save_json",
    "save_json",
    "save_pydantic",
    "save_pydantic",
    "save_toml",
    "save_toml",
    "save_yaml",
    "save_yaml",
    "saves",
    "saves_json",
    "saves_pydantic",
    "saves_toml",
    "saves_yaml",
    "serialize",
    "toml",
    "yaml",
]



"""This module provides timing utilities including classes and functions for measuring and recording execution time."""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)



from collections.abc import Iterator
from typing import Protocol, runtime_checkable


@runtime_checkable
class SizedIterable[T](Protocol):
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[T]: ...



import atexit
import collections
import statistics
import textwrap
from collections.abc import Generator, Mapping, Sequence
from typing import overload

import attrs
from loguru import logger

from liblaf import grapes

from ._time import TimerName, get_time


@attrs.define
class TimerConfig:
    label: str | None = attrs.field(default=None, kw_only=True)
    timers: Sequence[TimerName | str] = attrs.field(
        factory=lambda: ["perf"], kw_only=True, on_setattr=attrs.setters.frozen
    )

    @property
    def default_timer(self) -> str:
        return self.timers[0]


@attrs.define
class BaseTimer(TimerConfig):
    _time_start: dict[str, float] = attrs.field(
        init=False, factory=dict, on_setattr=attrs.setters.frozen
    )
    _time_end: dict[str, float] = attrs.field(
        init=False, factory=dict, on_setattr=attrs.setters.frozen
    )

    def current_time(self, timer: str | None = None) -> float:
        timer = timer or self.default_timer
        return self._time_end[timer] - self._time_start[timer]

    def _start(self) -> None:
        for timer in self.timers:
            self._time_start[timer] = get_time(timer)

    def _end(self) -> None:
        for timer in self.timers:
            self._time_end[timer] = get_time(timer)

    @property
    def _current_record(self) -> Mapping[str, float]:
        return {
            timer: self._time_end[timer] - self._time_start[timer]
            for timer in self.timers
        }


@attrs.define
class TimerRecords(BaseTimer):
    depth: int = attrs.field(default=0, kw_only=True)
    log_level_record: int | str | None = attrs.field(default="DEBUG", kw_only=True)
    log_level_summary: int | str | None = attrs.field(default="INFO", kw_only=True)
    log_summary_at_exit: bool = attrs.field(default=False, kw_only=True)
    _records: dict[str, list[float]] = attrs.field(
        init=False,
        factory=lambda: collections.defaultdict(list),
        on_setattr=attrs.setters.frozen,
    )

    def __post_attrs_init__(self) -> None:
        if self.log_summary_at_exit:
            register_timer_at_exit(self)

    @overload
    def __getitem__(self, index: int) -> Mapping[str, float]: ...
    @overload
    def __getitem__(self, index: str) -> Sequence[float]: ...
    def __getitem__(self, index: int | str) -> Mapping[str, float] | Sequence[float]:
        if isinstance(index, int):
            return self.row(index)
        return self.column(index)

    def __len__(self) -> int:
        return self.count

    @property
    def columns(self) -> Sequence[str]:
        return self.timers

    @property
    def count(self) -> int:
        return self.n_rows

    @property
    def n_columns(self) -> int:
        return len(self.timers)

    @property
    def n_rows(self) -> int:
        return len(self.column())

    def column(self, timer: str | None = None) -> Sequence[float]:
        timer = timer or self.default_timer
        return self._records[timer]

    def human_record(self, index: int = -1, label: str | None = None) -> str:
        label = label or self.label or "Timer"
        text: str = f"{label} > "
        items: list[str] = []
        for timer, value in self.row(index).items():
            human_duration: str = grapes.human_duration(value)
            items.append(f"{timer}: {human_duration}")
        text += ", ".join(items)
        return text

    def human_summary(self, label: str | None = None) -> str:
        label = label or self.label or "Timer"
        header: str = f"{label} (total: {self.n_rows})"
        if self.n_rows == 0:
            return header
        body: str = ""
        for timer in self.columns:
            body += f"{timer} > "
            human_mean: str = grapes.human_duration_series(self.column(timer))
            human_median: str = grapes.human_duration(self.median(timer))
            body += f"mean: {human_mean}, median: {human_median}\n"
        body = body.strip()
        summary: str = header + "\n" + textwrap.indent(body, "  ")
        return summary

    def iter_columns(self) -> Generator[tuple[str, Sequence[float]]]:
        yield from self._records.items()

    def iter_rows(self) -> Generator[Mapping[str, float]]:
        for index in range(self.n_rows):
            yield self.row(index)

    def log_record(
        self,
        index: int = -1,
        label: str | None = None,
        depth: int = 1,
        level: int | str | None = None,
    ) -> None:
        level = level or self.log_level_record
        if level is None:
            return
        logger.opt(depth=self.depth + depth).log(
            level, self.human_record(index=index, label=label)
        )

    def log_summary(
        self, label: str | None = None, depth: int = 1, level: int | str | None = None
    ) -> None:
        level = level or self.log_level_summary
        if level is None:
            return
        logger.opt(depth=self.depth + depth).log(level, self.human_summary(label=label))

    def row(self, index: int) -> Mapping[str, float]:
        return {timer: values[index] for timer, values in self._records.items()}

    # region statistics

    def max(self, timer: str | None = None) -> float:
        return max(self.column(timer))

    def mean(self, timer: str | None = None) -> float:
        return statistics.mean(self.column(timer))

    def median(self, timer: str | None = None) -> float:
        return statistics.median(self.column(timer))

    def min(self, timer: str | None = None) -> float:
        return min(self.column(timer))

    def std(self, timer: str | None = None) -> float:
        return statistics.stdev(self.column(timer))

    # endregion statistics

    def _append(
        self, seconds: Mapping[str, float] = {}, nanoseconds: Mapping[str, float] = {}
    ) -> None:
        for key, value in seconds.items():
            self._records[key].append(value)
        for key, value in nanoseconds.items():
            self._records[key].append(value * 1e-9)

    def _end(self) -> None:
        super()._end()
        self._append(seconds=self._current_record)


TIMERS: list[TimerRecords] = []


def register_timer_at_exit(timer: TimerRecords) -> None:
    TIMERS.append(timer)


def log_summary() -> None:
    for timer in TIMERS:
        if timer.log_summary_at_exit and timer.n_rows > 1:
            timer.log_summary()


atexit.register(log_summary)



import io
import os
from pathlib import Path
from typing import Any, override

from ruamel.yaml import YAML

from liblaf import grapes

from ._abc import AbstractSerializer


class YAMLSerializer(AbstractSerializer):
    @override
    def load(self, fpath: str | os.PathLike[str], **kwargs) -> Any:
        fpath: Path = grapes.as_path(fpath)
        yaml = YAML(**kwargs)
        with fpath.open() as fp:
            return yaml.load(fp)

    @override
    def loads(self, data: str, **kwargs) -> Any:
        stream = io.StringIO(data)
        yaml = YAML(**kwargs)
        return yaml.load(stream)

    @override
    def save(self, fpath: str | os.PathLike[str], data: Any, **kwargs) -> None:
        fpath: Path = grapes.as_path(fpath)
        yaml = YAML(**kwargs)
        with fpath.open("w") as fp:
            yaml.dump(data, fp)

    @override
    def saves(self, data: Any, **kwargs) -> str:
        stream = io.StringIO()
        yaml = YAML(**kwargs)
        yaml.dump(data, stream)
        return stream.getvalue()


yaml = YAMLSerializer()
load_yaml = yaml.load
loads_yaml = yaml.loads
save_yaml = yaml.save
saves_yaml = yaml.saves



from liblaf import grapes

with grapes.optional_imports(extra="duration"):
    import about_time


def human_throughout(value: float, unit: str = "", prec: int | None = None) -> str:
    """Convert a numeric value representing throughput into a human-readable string.

    Args:
        value: The numeric value of the throughput.
        unit: The unit of the throughput (e.g., "it/s", "B/s").
        prec: The precision of the human-readable output. If `None`, default precision is used.

    Returns:
        A human-readable string representing the throughput.
    """
    # TODO: remove dependency on `about-time`
    ht = about_time.HumanThroughput(value, unit)
    return ht.as_human(prec)



import os
from pathlib import Path
from typing import Any, override

import tomlkit

from liblaf import grapes

from ._abc import AbstractSerializer


class TOMLSerializer(AbstractSerializer):
    @override
    def load(self, fpath: str | os.PathLike[str], **kwargs) -> tomlkit.TOMLDocument:
        fpath: Path = grapes.as_path(fpath)
        with fpath.open() as fp:
            return tomlkit.load(fp, **kwargs)

    @override
    def loads(self, data: str, **kwargs) -> tomlkit.TOMLDocument:
        return tomlkit.loads(data, **kwargs)

    @override
    def save(self, fpath: str | os.PathLike[str], data: Any, **kwargs) -> None:
        fpath: Path = grapes.as_path(fpath)
        with fpath.open("w") as fp:
            tomlkit.dump(data, fp, **kwargs)

    @override
    def saves(self, data: Any, **kwargs) -> str:
        return tomlkit.dumps(data, **kwargs)


toml = TOMLSerializer()
load_toml = toml.load
loads_toml = toml.loads
save_toml = toml.save
saves_toml = toml.saves



from ._base import (
    BaseTimer,
    TimerConfig,
    TimerRecords,
    register_timer_at_exit,
)
from ._function import TimedFunction
from ._iterable import TimedIterable
from ._time import TimerName, get_time
from ._timer import Timer, timer

__all__ = [
    "BaseTimer",
    "TimedFunction",
    "TimedIterable",
    "Timer",
    "TimerConfig",
    "TimerName",
    "TimerRecords",
    "get_time",
    "register_timer_at_exit",
    "timer",
]



import os
import tempfile
from pathlib import Path
from typing import Any

from liblaf import grapes


class AbstractSerializer:
    def load(self, fpath: str | os.PathLike[str], **kwargs) -> Any:
        if type(self).loads is not AbstractSerializer.loads:
            fpath: Path = grapes.as_path(fpath)
            return self.loads(fpath.read_text(), **kwargs)
        raise NotImplementedError

    def loads(self, data: str, **kwargs) -> Any:
        if type(self).load is not AbstractSerializer.load:
            with tempfile.NamedTemporaryFile("w") as fp:
                fp.write(data)
                return self.load(Path(fp.name), **kwargs)
        raise NotImplementedError

    def save(self, fpath: str | os.PathLike[str], data: Any, **kwargs) -> None:
        if type(self).saves is not AbstractSerializer.saves:
            fpath: Path = grapes.as_path(fpath)
            fpath.write_text(self.saves(data, **kwargs))
        raise NotImplementedError

    def saves(self, data: Any, **kwargs) -> str:
        if type(self).save is not AbstractSerializer.save:
            with tempfile.NamedTemporaryFile("w") as fp:
                self.save(Path(fp.name), data, **kwargs)
                return fp.read()
        raise NotImplementedError



import os
from typing import Any

import pydantic

from ._serde import load, loads, save, saves


def load_pydantic[C: pydantic.BaseModel](
    fpath: str | os.PathLike[str], cls: type[C], *, ext: str | None = None, **kwargs
) -> C:
    data: Any = load(fpath, ext=ext, **kwargs)
    return cls.model_validate(data)


def loads_pydantic[C: pydantic.BaseModel](
    data: str, cls: type[C], *, ext: str | None = None, **kwargs
) -> C:
    data: Any = loads(data, ext=ext, **kwargs)
    return cls.model_validate(data)


def save_pydantic(
    fpath: str | os.PathLike[str],
    data: pydantic.BaseModel,
    *,
    ext: str | None = None,
    # pydantic.BaseModel.model_dump(**kwargs)
    context: Any | None = None,
    by_alias: bool = False,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
    exclude_none: bool = False,
    round_trip: bool = False,
    serialize_as_any: bool = False,
) -> None:
    save(
        fpath,
        data.model_dump(
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            serialize_as_any=serialize_as_any,
        ),
        ext=ext,
    )


def saves_pydantic(
    data: pydantic.BaseModel,
    *,
    ext: str | None = None,
    # pydantic.BaseModel.model_dump(**kwargs)
    context: Any | None = None,
    by_alias: bool = False,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
    exclude_none: bool = False,
    round_trip: bool = False,
    serialize_as_any: bool = False,
) -> str:
    return saves(
        data.model_dump(
            context=context,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            serialize_as_any=serialize_as_any,
        ),
        ext=ext,
    )



import logging


def clear_handlers() -> None:
    for logger in logging.root.manager.loggerDict.values():
        if isinstance(logger, logging.PlaceHolder):
            continue
        logger.handlers.clear()
        logger.propagate = True



from collections.abc import Iterable, Iterator

import attrs

from ._base import TimerRecords


@attrs.define
class TimedIterable[T](TimerRecords):
    total: int | None = attrs.field(default=None, kw_only=True)
    _iterable: Iterable[T] = attrs.field(
        alias="iterable", on_setattr=attrs.setters.frozen
    )

    def __attrs_post_init__(self) -> None:
        self.label = self.label or "Iterable"

    def __contains__(self, x: object, /) -> bool:
        return x in self._iterable  # pyright: ignore[reportOperatorIssue]

    def __iter__(self) -> Iterator[T]:
        for item in self._iterable:
            self._start()
            yield item
            self._end()
            self.log_record(depth=2)
        self.log_summary(depth=2)

    def __len__(self) -> int:
        if self.total is not None:
            return self.total
        return len(self._iterable)  # pyright: ignore[reportArgumentType]



import functools
from typing import Any

from loguru import logger
from typing_extensions import deprecated

# https://stackoverflow.com/a/66062313


@deprecated("Use `filter_once()` + `logger.bind(once=True).trace(...)` instead.")
@functools.lru_cache
def trace_once(message: Any, *args, **kwargs) -> None:
    logger.trace(message, *args, **kwargs)


@deprecated("Use `filter_once()` + `logger.bind(once=True).debug(...)` instead.")
@functools.lru_cache
def debug_once(message: Any, *args, **kwargs) -> None:
    logger.debug(message, *args, **kwargs)


@deprecated("Use `filter_once()` + `logger.bind(once=True).info(...)` instead.")
@functools.lru_cache
def info_once(message: Any, *args, **kwargs) -> None:
    logger.info(message, *args, **kwargs)


@deprecated("Use `filter_once()` + `logger.bind(once=True).success(...)` instead.")
@functools.lru_cache
def success_once(message: Any, *args, **kwargs) -> None:
    logger.success(message, *args, **kwargs)


@deprecated("Use `filter_once()` + `logger.bind(once=True).warning(...)` instead.")
@functools.lru_cache
def warning_once(message: Any, *args, **kwargs) -> None:
    logger.warning(message, *args, **kwargs)


@deprecated("Use `filter_once()` + `logger.bind(once=True).error(...)` instead.")
@functools.lru_cache
def error_once(message: Any, *args, **kwargs) -> None:
    logger.error(message, *args, **kwargs)


@deprecated("Use `filter_once()` + `logger.bind(once=True).critical(...)` instead.")
@functools.lru_cache
def critical_once(message: Any, *args, **kwargs) -> None:
    logger.critical(message, *args, **kwargs)


@deprecated("Use `filter_once()` + `logger.bind(once=True).log(...)` instead.")
@functools.lru_cache
def log_once(level: int | str, message: Any, *args, **kwargs) -> None:
    logger.log(level, message, *args, **kwargs)


@deprecated("Use `filter_once()` + `logger.bind(once=True).exception(...)` instead.")
@functools.lru_cache
def exception_once(message: Any, *args, **kwargs) -> None:
    logger.exception(message, *args, **kwargs)
